﻿Perception = LibStub("AceAddon-3.0"):NewAddon("Perception", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0");
local AceGUI = LibStub("AceGUI-3.0")
local LibGraph = LibStub("LibGraph-2.0-ArcRev")
--[[
	Author: Arclight-The Sha'tar EU
	Created: 2012-02-09
	
	
	Stream data layout:
	Streams = 
	{
		["Arclight"] =
		{
			["mana"] =
			{
				Raw = { {x,y}, {x2,y2} }
				Show = true
				Type = "line" --graph type (line/story)
				Color = {r,g,b,a}
				Recording = true				
				Capture = f --function called when a timer update happens				
			}
		}
	}
]]

--cached local funtions
local GetTime = GetTime
local pairs = pairs

--timing
local LastUpdate = GetTime()
local ViewSpanOffset = 0

--frame and events list for subscriptions
local frame, events = nil, {};
--
local defaults = {
			profile = {
				CollectionInterval = 0.7,
				ViewSpan = 30,
				GlobalCollection = true,
				LiveUpdate = true,
				ShowMainWindow = true
			}
		}

local options = {
    name = "Perception",
    handler = Perception,
    type = "group",
    args = {
        interval = {
            type = "range",			
			min = 0.2,
			max = 2,
			step = 0.05,
			bigStep = 0.1,
            name = "Data Collection Interval",
            desc = "The frequency of data collection in seconds (smaller numbers = higher frequency, but also higher CPU and memory usage)",            
            get = "GetCollectionInterval",
            set = "SetCollectionInterval",
			width = "double"
        },
        viewspan = {
            type = "range",
			min = 5,
			max = 300, -- 5 minutes - quite a lot for a window
			step = 1,
			bigStep = 5,
            name = "View Span",
            desc = "Number of seconds you want to see in the graph frame",
            get = "GetViewSpan",
            set = "SetViewSpan",
			width = "double"
        },
        collection = {
            type = "toggle",			
            name = "Global Collection",
            desc = "Toggles the recording of data on or off",
            get = "IsCollecting",
            set = "ToggleCollecting",
			order = 0,
			width = "full"
        },
		live = {
			type = "toggle",			
			name = "Live Graph Updates",
			desc = "Toggle live updating of the graph",
			set = "ToggleLiveUpdate",
			get = "GetLiveUpdate",
			width = "full",
			order = 1
		},
		showstream = {
			type = "input",
			guiHidden = true,
			name = "Toggle a stream",
			desc = "Show/hide a stream. Useage: /perc showstream [player] [streamtype]",
			set = "ConsoleShowStream"
		},
		reset = {
			type = "input",
			guiHidden = true,
			name = "Reset all data",
			desc = "Reset all data collected by Perception",
			set = "ConsoleReset"
		},
		show = {
			type = "toggle",
			name = "Show Main Window",
			desc = "Show or hide the main graph window",
			set = "ToggleShowMainWindow",
			get = "IsMainWindowVisible",
			width = "full",
			order = 2
		}
    },
}

local Colors = {["Class"]={
					["HUNTER"] = { 0.67, 0.83, 0.45, 1 },
					["WARLOCK"] = { 0.58, 0.51, 0.79, 1 },
					["PRIEST"] = { 1.0, 1.0, 1.0, 1 },
					["PALADIN"] = { 0.96, 0.55, 0.73, 1 },
					["MAGE"] = { 0.41, 0.8, 0.94, 1 },
					["ROGUE"] = { 1.0, 0.96, 0.41, 1 },
					["DRUID"] = { 1.0, 0.49, 0.04, 1 },
					["SHAMAN"] = { 0.14, 0.35, 1.0, 1 },
					["WARRIOR"] = { 0.78, 0.61, 0.43, 1 },
					["DEATHKNIGHT"] = { 0.77, 0.12, 0.23, 1 }
					}
				}

function Perception:GetCollectionInterval(info)
	return self.db.profile.CollectionInterval
end
function Perception:SetCollectionInterval(info, newval)
	self.db.profile.CollectionInterval = newval
	if self.dataTimer then
		self:CancelTimer(self.dataTimer)
	end
	self.dataTimer = self:ScheduleRepeatingTimer("OnDataTimer", self.db.profile.CollectionInterval)
	self:Trace("Collection Interval set to "..newval)
end

function Perception:GetViewSpan(info)
	return self.db.profile.ViewSpan
end
function Perception:SetViewSpan(info, newval)
	self.db.profile.ViewSpan = newval
	self:Trace("View Span set to "..newval)
end

function Perception:IsCollecting(info)
	return self.db.profile.GlobalCollection
end
function Perception:ToggleCollecting(info, newval)
	self.db.profile.GlobalCollection = not self.db.profile.GlobalCollection
	if (self.db.profile.GlobalCollection) then
		self:Print("Global collection is ON")
	else
		self:Print("Global collection is OFF")
	end
end

function Perception:ToggleShowMainWindow()
	if self.MainWindow then
		self.db.profile.ShowMainWindow = not self.db.profile.ShowMainWindow
		if self.db.profile.ShowMainWindow then
			self.MainWindow:Show()
		else
			self.MainWindow:Hide()
		end
	end
end
function Perception:IsMainWindowVisible()
	return self.db.profile.ShowMainWindow
end

local Modules = {} --tracking modules
local Streams = {}

function Perception:ClassColor(playername)
	local _, class = UnitClass(playername)	
	if class then 
		return Colors["Class"][class]
	end
	return {1,1,1,0.5}
end

function Perception:DefaultStream(color)
	local c = color or {1,1,1,1}
	return { Raw = {}, Show = false, Color = c, Capture = nil, Recording = true, Type = "line" }
end

function Perception:RegisterStream(player, type, func, color, graphtype)
	self:Trace(string.format("Perception:RegisterStream(%s, %s)", player, type))

	Streams[player] = Streams[player] or {}
	Streams[player][type] = Streams[player][type] or self:DefaultStream()
	local stream = Streams[player][type]
	stream.Capture = func
	stream.Color = color or self:ClassColor(player)
	stream.Type = graphtype or "line"
	stream.Recording = true

	--auto showing stream for lols
	self:ShowStream(player, type, true)

	return stream
end

function Perception:RegisterModule(name, module)
	self:Trace("Perception:RegisterModule("..name..")")
	Modules[name] = module
end

function Perception:ShowStream(player, type, show)
	if show then
		self:Trace("ShowStream("..player..", "..type..", ON)")
	else
		self:Trace("ShowStream("..player..", "..type..", OFF)")
	end
	if ( Streams[player] and Streams[player][type] ) then
		Streams[player][type].Show = show
		if ( self.MainWindowGraph ) then
			local g = self.MainWindowGraph
			local r = Streams[player][type]
			if show then
				if r.Type == "line" then
					g:AddDataSeries(r.Raw, r.Color)
				elseif r.Type == "story" then
					g:AddStorySeries(r.Raw)
				end
			else
				g:RemoveDataSeries(r.Raw)
				g:RemoveStorySeries(r.Raw)
			end 
		end
	end
end

function Perception:ConsoleShowStream(info, input)
	local p, t = strsplit(" ", input)
	if ( Streams[p] and Streams[p][t] ) then
		local b = Streams[p][t].Show		
		self:ShowStream(p, t, not b)
	end
end

function Perception:ConsoleReset(info, input)
	for u,us in pairs(Streams) do
		for t,s in pairs(us) do
			local r = s.Raw
			for k,v in pairs(r) do
				r[k] = nil
			end
		end
	end
	self:Print("Data reset")
end

function Perception:ToggleLiveUpdate(info)
	self.db.profile.LiveUpdate = not self.db.profile.LiveUpdate
end
function Perception:GetLiveUpdate()
	return self.db.profile.LiveUpdate
end

function Perception:IsRecording(unit, streamtype)
	
	if Perception:IsCollecting() == false then
		return false
	end

	if Streams[unit] and Streams[unit][streamtype] then
		return Streams[unit][streamtype].Recording
	end
	
	return false	
end

local function DebugPrintTable(t)
	for k,v in pairs(t) do
		if type(v) ~= "table" then
			Perception:Trace(k .. ":" .. v)
		end
	end
end

function Perception:AddStoryPoint(unit, streamtype, x, y, spellid, tooltip)
	if Streams[unit] and Streams[unit][streamtype] then
		local point = {x=x, y=y, spellid=spellid, tooltip=tooltip}
		--DebugPrintTable(point)
		tinsert(Streams[unit][streamtype].Raw, point)
	end
end

function Perception:AddDataPoint(unit, streamtype, x, y)
	if Streams[unit] and Streams[unit][streamtype] then
		tinsert(Streams[unit][streamtype].Raw, {x, y})
	end
end

function Perception:PARTY_MEMBERS_CHANGED(event, ...)

	self:Trace("PARTY_MEMBERS_CHANGED")
	
	local currentmembers = { } 
	currentmembers[UnitName("player")] = true--start with the player cos he/she is obviously here

	local nRaid = GetNumRaidMembers()
	local nParty = GetNumPartyMembers()
	if nRaid > 0 then
		for i = 1, nRaid do
			local unitname = UnitName("raid"..i)
			currentmembers[unitname] = true

			--notify registered modules that a new player has arrived
			for n, m in pairs(Modules) do
				m:NewPlayer(unitname)
			end
			
		end
	elseif GetNumPartyMembers() > 0 then
		for i = 1, nParty do
			local unitname = UnitName("party"..i)
			currentmembers[unitname] = true
			--notify registered modules that a new player has arrived
			for n, m in pairs(Modules) do
				m:NewPlayer(unitname)
			end			
		end
	end

	for k,v in pairs(Streams) do
		if not currentmembers[k] then
			for k2, s in pairs(v) do
				if s.Recording then
					self:Trace(k .. " is not longer in the group. Turning recording OFF for ".. k)
					s.Recording = false
				end
			end
		end
	end
end

function Perception:PLAYER_ENTERING_WORLD(event, ...)
	--self:Print("PLAYER_ENTERING_WORLD")

	local playername = UnitName("player")
	for n, m in pairs(Modules) do
		m:NewPlayer(playername)
	end
end

function Perception:OnDataTimer()
	if not self:IsCollecting() then
		return
	end
	
	local t = GetTime()

	for unitname, v  in pairs(Streams) do		
		for streamtype, stream in pairs(v) do
			if stream.Recording and stream.Capture then				
				stream:Capture(unitname, streamtype, t) -- call the function in the stream so it can capture its own data
			end
		end
	end

end

function Perception:UpdateGraphAxis()
	if ( Perception.MainWindow and Perception.MainWindowGraph ) then
		local g = Perception.MainWindowGraph
		local t = GetTime()
		local delta = LastUpdate - t
		LastUpdate = t
		g:SetXAxis(t - Perception.db.profile.ViewSpan - ViewSpanOffset, t - ViewSpanOffset)
	end
end

local function OnUpdate()	
	if ( Perception.MainWindow and Perception.MainWindowGraph ) then		
		local m, g = Perception.MainWindow, Perception.MainWindowGraph
		local w, h = m:GetWidth(), m:GetHeight()
		if ( g:GetWidth() ~= w - g.internalpadding ) then
			g:SetWidth(w - g.internalpadding)
			g:RefreshGraph()
		end
		if ( g:GetHeight() ~= h - g.internalpadding ) then
			g:SetHeight(h - g.internalpadding)
			g:RefreshGraph()
		end

		if Perception.db.profile.LiveUpdate then
			Perception.UpdateGraphAxis()	
		end
	end
end

function Perception:SetupEvents()	
	self:RegisterEvent("PARTY_MEMBERS_CHANGED")
	self:RegisterEvent("PLAYER_ENTERING_WORLD")
	frame = CreateFrame("Frame")
	frame:SetScript("OnUpdate", OnUpdate);
end

local function OnMouseWheel(frame, delta)
	local vs = Perception.db.profile.ViewSpan
	frame:StopMovingOrSizing()
	if IsShiftKeyDown() then
		vs = vs + (delta*2)
		if vs > 5 and vs < 300 then
			Perception.db.profile.ViewSpan = vs
			Perception.UpdateGraphAxis()
		end
	else
		local factor = Perception.db.profile.ViewSpan / 15
		delta = delta * factor
		ViewSpanOffset = math.max(ViewSpanOffset + delta,0)
		Perception.UpdateGraphAxis()
	end
end

function Perception:CreateMainWindow()
	
	local f=CreateFrame("Frame","Perception_MainWindow",UIParent)
	f.Me=me

	f:ClearAllPoints()
	f:SetPoint("CENTER",UIParent)
	f:SetHeight(432+20)
	f:SetWidth(625+24)

	
	f:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background", tile = true, tileSize = 16,
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border", edgeSize = 16, 
		insets = { left = 4, right = 4, top = 4, bottom = 4}
	})
	
	f:SetBackdropBorderColor(0.2,0.2,0.2)
	f:SetBackdropColor(24/255, 24/255, 24/255, 0.2)
	
	f:EnableMouse(true)
	f:SetMovable(true)
	f:SetResizable(true)
	f:SetMinResize(200,100)
	f:RegisterForDrag("LeftButton")
	f:SetScript("OnDragStart",f.StartMoving)
	f:SetScript("OnDragStop",f.StopMovingOrSizing)
	f:SetScript("OnMouseWheel", OnMouseWheel)

	f.CloseButton=CreateFrame("Button",nil,f)
	f.CloseButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up.blp")
	f.CloseButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down.blp")
	f.CloseButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp")
	f.CloseButton:SetWidth(20)
	f.CloseButton:SetHeight(20)
	f.CloseButton:SetPoint("TOPRIGHT",f,"TOPRIGHT",-4,-4)
	f.CloseButton:SetScript("OnClick",function(this) 
								--this:GetParent():Hide(); 
								Perception:ToggleShowMainWindow(); end)

	f.DragBottomRight = CreateFrame("Button", nil, f)
	f.DragBottomRight:Show()
	f.DragBottomRight:SetFrameLevel( f:GetFrameLevel() + 10)
	f.DragBottomRight:SetNormalTexture("Interface\\AddOns\\Recount\\textures\\ResizeGripRight")
	f.DragBottomRight:SetHighlightTexture("Interface\\AddOns\\Recount\\textures\\ResizeGripRight")
	f.DragBottomRight:SetWidth(16)
	f.DragBottomRight:SetHeight(16)
	f.DragBottomRight:SetPoint("BOTTOMRIGHT", f, "BOTTOMRIGHT", 0, 0)
	f.DragBottomRight:EnableMouse(true)
	f.DragBottomRight:SetScript("OnMouseDown", function(self,button) if ((( not self:GetParent().isLocked ) or ( self:GetParent().isLocked == 0 ) ) and ( button == "LeftButton" ) ) then self:GetParent().isResizing = true; self:GetParent():StartSizing("BOTTOMRIGHT") end end )
	f.DragBottomRight:SetScript("OnMouseUp", function(self,button) if self:GetParent().isResizing == true then self:GetParent():StopMovingOrSizing(); self:GetParent().isResizing = false; end end )

	local w, h = f:GetWidth(), f:GetHeight()
	local internalpadding = 10
	local g = LibGraph:CreateGraphLine("TestLineGraph",f,"CENTER","CENTER",0,0,w-internalpadding,h-internalpadding)		
	g.internalpadding = internalpadding
	g:SetXAxis(0,self.db.profile.ViewSpan)
	g:SetYAxis(0,105)
	g:SetGridSpacing(self.db.profile.ViewSpan / 2,0) -- 2 grid lines every ViewSpan seconds
	g:SetGridColor({0.5,0.5,0.5,0.2})
	g:SetAxisDrawing(false,false)
	g:SetAxisColor({1.0,1.0,1.0,1.0})
	g:SetAutoScale(false)

	self.MainWindow = f
	self.MainWindowGraph = g
end


function Perception:SetupOptions()
	self.db = LibStub("AceDB-3.0"):New("PerceptionDB", defaults, "Defaults")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Perception", options, {"perception", "perc"})
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("Perception", "Perception")
end

function Perception:OnInitialize()
    -- Called when the addon is loaded
	self:SetupEvents()
	self:SetupOptions()
	self:CreateMainWindow()	
	self.dataTimer = self:ScheduleRepeatingTimer("OnDataTimer", self.db.profile.CollectionInterval)
end

function Perception:OnEnable()
    -- Called when the addon is enabled
end

function Perception:OnDisable()
    -- Called when the addon is disabled
end
